*-----------------------------------------------------------------------
*-- File........: Printer.prg
*-- Programmer..: Bowen Moursund (CIS: 72662,436), Copyright 1997
*-----------------------------------------------------------------------

CLASS PrinterProperties(cPrinterDriver, cPrinterName, cPrinterPort)
*-----------------------------------------------------------------------
*-- Programmers.: Bowen Moursund (CIS: 72662,436)
*--               Jim Sare (CIS: 71562,413)
*-- Date........: 02/28/97
*-- Written for.: Visual dBASE v5.5
*-- Rev. History: 04/01/96 - Original version
*--               08/28/96 - New VdBDevMode class
*--               02/28/97 - Init_pDriver method added
*--                        - Internationalization
*--               04/29/97 - v2.3a, bug fixes: on error
*-----------------------------------------------------------------------

    #include Printer.H

    protect pInitDriver

    set procedure to ErrorMsg.Prg additive
    this.Err = new ErrorMessage(new PrnPropsErrors())

    *-- list of installed printer drivers
    *-- from Win.Ini [Devices] section
    this.Printers = new PrinterProfiles()
    this.LastSpecific = this.Printers.ListString(this.Printers.Printers[1,1],;
                                                 this.Printers.Printers[1,3])

    this.DefaultIsSelected = .t.
    store "" to this.DefaultPrinterName,;
                this.DefaultPrinterDriver,;
                this.DefaultPrinterPort,;
                this.PrinterDriver,;
                this.PrinterName,;
                this.PrinterPort,;
                this.CurrentDriver

    this.PaperName = ""  && will hold name of selected paper size
    this.BinName = ""  && will hold name of selected paper bin

    *-- supported DEVMODE fields
    store 0 to this.PaperSize,;
               this.DefaultSource,;
               this.Orientation

    this.DevMode = .f.  && will hold VdBDevMode structure obj reference
    this.DevModeParm = ""  && will hold DEVMODE string
    this.PaperSizes = .f.  && will hold array of paper size values
    this.PaperNames = .f.  && will hold array of paper names
    this.DefaultSources = .f.  && will hold array of paper source values
    this.BinNames = .f.  && will hold array of paper source names
    this.nPaperSizes = 0  && will hold number of paper sizes
    this.nPaperNames = 0
    this.nBins = 0  && will hold number of paper sources
    this.nBinNames = 0
    this.Landscape = .f.  && driver supports page orientation

    if pcount() = 0
        class::Reset()
    else
        if pcount() = 3 .and. class::AreValidDriverParms(cPrinterDriver,;
                                                         cPrinterName,;
                                                         cPrinterPort)
            local oPrinter
            oPrinter = class::GetDefaultPrinter()
            this.DefaultPrinterName  = oPrinter.DefaultPrinterName
            this.DefaultPrinterDriver= oPrinter.DefaultPrinterDriver
            this.DefaultPrinterPort  = oPrinter.DefaultPrinterPort
            this.PrinterDriver = cPrinterDriver
            this.PrinterName = cPrinterName
            this.PrinterPort = cPrinterPort
            class::pInitDriver()
        else
            class::Error(3, "")
            class::Reset()
        endif
    endif


    PROCEDURE AreValidDriverParms
    parameters cPrinterDriver, cPrinterName, cPrinterPort
    RETURN type("cPrinterDriver") == "C" .and.;
           type("cPrinterName") == "C" .and.;
           type("cPrinterPort") == "C" .and.;
           "" <> cPrinterDriver .and.;
           "" <> cPrinterName .and.;
           "" <> cPrinterPort


    PROCEDURE Reset
    local oPrinter
    oPrinter = class::GetDefaultPrinter()
    this.DefaultPrinterName  = oPrinter.DefaultPrinterName
    this.DefaultPrinterDriver= oPrinter.DefaultPrinterDriver
    this.DefaultPrinterPort  = oPrinter.DefaultPrinterPort
    this.PrinterName  = oPrinter.DefaultPrinterName
    this.PrinterDriver= oPrinter.DefaultPrinterDriver
    this.PrinterPort  = oPrinter.DefaultPrinterPort
    class::pInitDriver()
    this.DefaultIsSelected = .t.


    PROCEDURE InitDriver(cPrinterDriver, cPrinterName, cPrinterPort)
    if class::AreValidDriverParms(cPrinterDriver, cPrinterName, cPrinterPort)
        this.PrinterDriver = cPrinterDriver
        this.PrinterName = cPrinterName
        this.PrinterPort = cPrinterPort
        class::pInitDriver()
    endif

    PROCEDURE Init_pDriver
    *-- Init VdB settings. Sets only driver and page orientation
    local cPrinterDriver, cPrinterName, cPrinterPort, nComma
    nComma = at(",", _pdriver)
    cPrinterDriver = ltrim(trim(left(_pdriver, nComma-1)))
    cPrinterName = ltrim(trim(substr(_pdriver, nComma+1)))
    cPrinterPort = setto("printer")
    if class::AreValidDriverParms(cPrinterDriver, cPrinterName, cPrinterPort)
        this.PrinterDriver = cPrinterDriver
        this.PrinterName = cPrinterName
        this.PrinterPort = cPrinterPort
        class::pInitDriver()
        if _porientation = "PORTRAIT"
            form.PProps.Orientation = DMORIENT_PORTRAIT
        else
            form.PProps.Orientation = DMORIENT_LANDSCAPE
        endif
    endif

    PROCEDURE pInitDriver  && protected
    if "" <> this.CurrentDriver
        private cCurrentDriver
        cCurrentDriver = this.CurrentDriver
        release DLL &cCurrentDriver.
    endif

    *-- load new driver
    private cDriverFile
    cDriverFile = this.PrinterDriver+".DRV"
    private lSuccess, cOnError
    lSuccess = .t.
    cOnError = setto("on error")
    on error lSuccess = .f.
    load DLL &cDriverFile.
    if .not. isblank(cOnError)
        on error &cOnError.
    else
        on error
    endif
    if lSuccess
        *-- prototype printer driver functions
        extern CINT ExtDeviceMode(CHANDLE,CHANDLE,CPTR,CSTRING,;
                                  CSTRING,CPTR,CLONG,CINT) &cDriverFile.
        extern CLONG DeviceCapabilities(CSTRING,CSTRING,CWORD,;
                                        CPTR,CPTR) &cDriverFile.
        this.CurrentDriver = this.PrinterDriver
    else
        class::Error(1, cDriverFile+": ")
        RETURN
    endif

    this.DefaultIsSelected = .f.
    this.DevMode = class::GetVdBDevMode()
    this.PaperSizes = class::GetPaperSizes()
    this.DefaultSources = class::GetDefaultSources()
    this.PaperNames = class::GetPaperNames()
    this.BinNames = class::GetBinNames()
    this.Landscape = class::DoesLandscape()

    this.nPaperSizes = this.PaperSizes.size
    this.nBins = this.DefaultSources.size
    this.nPaperNames = this.PaperNames.size
    this.nBinNames = this.BinNames.size

    this.Orientation = this.DevMode.dmOrientation()
    this.PaperSize = this.DevMode.dmPaperSize()
    this.DefaultSource = this.DevMode.dmDefaultSource()

    this.PaperName = iif(this.nPaperSizes > 0,;
                         class::GetPaperName(this.PaperSize), "")
    this.BinName = iif(this.nBins > 0,;
                       class::GetBinName(this.DefaultSource), "")


    PROCEDURE Release
    private cDriver, cOnError
    cDriver = this.PrinterDriver
    cOnError = setto("on error")
    on error *
    release DLL &cDriver.
    if .not. isblank(cOnError)
        on error &cOnError.
    else
        on error
    endif
    release object this


    PROCEDURE GetVdBDevMode
    *-- Get size of driver's DEVMODE structure
    local hDriver, nDevModeBytes
    PRIVATE oDevMode
    hDriver = GetModuleHandle(this.PrinterDriver)
    nDevModeBytes = ExtDeviceMode(0,hDriver,0,this.PrinterName,;
                                  this.PrinterPort,0,0,0)
    if nDevModeBytes > 0  && if a meaningful return
        oDevMode = this.DevMode
        If Type("oDevMode") # "O"
           oDevMode = new VdBDevMode()  && class defined below
        EndIf
        local cStructString
        cStructString = space(nDevModeBytes)
        if ExtDeviceMode(0,hDriver,cStructString,this.PrinterName,;
                         this.PrinterPort,0,0,DM_COPY) > 0
            *-- set this.DevMode structure to values
            *-- retrieved by ExtDeviceMode
            oDevMode.Structure(cStructString)
        else
            class::Error(2, this.PrinterDriver + ": ")
        endif
    else
        oDevMode = .f.  && stub out DevMode property
        class::Error(2, this.PrinterDriver + ": ")
    endif
    RETURN oDevMode
    *-- EoP: InitDevMode


    PROCEDURE SetVdBDevModeFields
    if type("this.DevMode") <> "O"  && no valid DevMode structure
        *-- set default value for PESelectprinter DEVMODE parm
        this.DevModeParm = ""
        RETURN
    endif

    *-- set page orientation field
    this.DevMode.dmOrientation(this.Orientation)

    *-- set paper size field
    this.DevMode.dmPaperSize(this.PaperSize)

    *-- set bin/source field
    this.DevMode.dmDefaultSource(this.DefaultSource)

    *-- copy the structure to a string to pass to CRPE
    this.DevModeParm = this.DevMode.Structure()
    *-- EoP: SetVdBDevModeFields


    PROCEDURE GetPaperNames
    local n, nPos, nNames, cBuffer, cName, aPaperNames, nPaperNames
    cBuffer = space(32000)
    nPaperNames =  DeviceCapabilities(this.PrinterName,;
                                      this.PrinterPort,;
                                      DC_PAPERNAMES,;
                                      cBuffer, 0)
    aPaperNames = new array(nPaperNames)
    if nPaperNames > 0
        nPos = 1
        for n = 1 to nPaperNames
            cName = substr(cBuffer, nPos)
            cName = left(cName, at(chr(0), cName) - 1)
            nPos = nPos + len(cName) + 1
            aPaperNames[n] = oem(ltrim(trim(cName)))
        endfor
    endif
    RETURN aPaperNames


    PROCEDURE GetPaperSizes
    local n, cBuffer, nPaperSize, aPaperSizes, nPaperSizes
    nPaperSizes = DeviceCapabilities(this.PrinterName, ;
                                 this.PrinterPort, ;
                                 DC_PAPERS, 0, 0)
    cBuffer = space(nPaperSizes*2)
    DeviceCapabilities(this.PrinterName, ;
                       this.PrinterPort, ;
                       DC_PAPERS, cBuffer, 0)
    aPaperSizes = new array(nPaperSizes)
    if nPaperSizes > 0
        for n = 1 to nPaperSizes
            aPaperSizes[n] = MAKEU16(cBuffer,(n*2)-1)
        endfor
    endif
    RETURN aPaperSizes


    PROCEDURE GetDefaultSources
    local n, cBuffer, nDefaultSource, aDefaultSources, nBins
    nBins = DeviceCapabilities(this.PrinterName, ;
                               this.PrinterPort, ;
                               DC_BINS, 0, 0)
    cBuffer = space(nBins*2)
    DeviceCapabilities(this.PrinterName, ;
                       this.PrinterPort, ;
                       DC_BINS, cBuffer, 0)
    aDefaultSources = new array(nBins)
    if nBins > 0
        for n = 1 to nBins
            aDefaultSources[n] = MAKEU16(cBuffer,(n*2)-1)
        endfor
    endif
    RETURN aDefaultSources


    PROCEDURE GetBinNames
    local n, nPos, cBuffer, cName, aBinNames, nBins
    cBuffer = space(32000)
    nBins =  DeviceCapabilities(this.PrinterName, ;
                                this.PrinterPort, ;
                                DC_BINNAMES, cBuffer, 0)
    aBinNames = new array(nBins)
    if nBins > 0
        nPos = 1
        for n = 1 to nBins
            cName = substr(cBuffer, nPos)
            cName = left(cName, at(chr(0), cName) - 1)
            nPos = nPos + len(cName) + 1
            aBinNames[n] = oem(ltrim(trim(cName)))
        endfor
    endif
    RETURN aBinNames


    PROCEDURE GetPaperSize(cPaperName)
    local nSubScript
    nSubScript = this.PaperNames.scan(cPaperName)
    RETURN iif(nSubScript > 0 .and. this.PaperSizes.size > 0,;
               this.PaperSizes[nSubScript], 0)


    PROCEDURE GetDefaultSource(cBinName)
    local nSubScript
    nSubScript = this.BinNames.scan(cBinName)
    RETURN iif(nSubScript > 0 .and. this.DefaultSources.size > 0,;
               this.DefaultSources[nSubScript], 0)


    PROCEDURE GetBinName(nDefaultSource)
    local nSubScript
    nSubScript = this.DefaultSources.scan(nDefaultSource)
    RETURN iif(nSubScript > 0 .and. this.BinNames.size > 0,;
               this.BinNames[nSubScript], "")


    PROCEDURE GetPaperName(nPaperSize)
    local nSubScript
    nSubScript = this.PaperSizes.scan(nPaperSize)
    RETURN iif(nSubScript > 0 .and. this.PaperNames.size > 0,;
               this.PaperNames[nSubScript], "")


    PROCEDURE DoesLandscape
    RETURN DeviceCapabilities(this.PrinterName, ;
                              this.PrinterPort, ;
                              DC_ORIENTATION, 0, 0) > 0


    PROCEDURE Copy
    local o
    o = new PrinterProperties(this.PrinterDriver,;
                              this.PrinterName,;
                              this.PrinterPort)
    o.DevModeParm = this.DevModeParm
    o.Orientation = o.DevMode.dmOrientation(this.Orientation)
    o.PaperSize = o.DevMode.dmPaperSize(this.PaperSize)
    o.DefaultSource = o.DevMode.dmDefaultSource(this.DefaultSource)

    o.PaperName = this.PaperName
    o.BinName = this.BinName
    o.CurrentDriver = this.CurrentDriver
    o.DefaultIsSelected = this.DefaultIsSelected
    o.LastSpecific = this.LastSpecific
    RETURN o


    PROCEDURE Error(nErrorCode, cMessagePrefix)
    this.Err.ErrorCode = nErrorCode
    this.Err.ShowErrMsg("PrinterProperties ", cMessagePrefix)


    PROCEDURE GetDefaultPrinter
    local oPrinter, cBuffer, nBytesUsed, nPos, cDefault
    if type("GetProfileString") <> "FP"
        extern CINT GetProfileString(CSTRING, CSTRING, CSTRING,;
                                     CPTR, CINT) KERNEL
    endif
    cBuffer = space(254)
    cDefault = resource(179,"CRPEWRAP.DLL")
    nBytesUsed = GetProfileString("windows", "device", cDefault,;
                                  cBuffer, 254)
    cBuffer = ltrim(trim(cBuffer))
    if right(cBuffer,1) = chr(0)
        cBuffer = left(cBuffer, len(cBuffer)-1)
    endif
    oPrinter = new object()
    if nBytesUsed = 0 .or. left(cBuffer, 9) = left(cDefault, 9)
        store "" to oPrinter.DefaultPrinterName,;
                    oPrinter.DefaultPrinterDriver,;
                    oPrinter.DefaultPrinterPort
    else
        nPos = at(",", cBuffer)
        oPrinter.DefaultPrinterName = trim(left(cBuffer, nPos - 1))
        cBuffer = ltrim(substr(cBuffer, nPos+1))
        nPos = at(",", cBuffer)
        oPrinter.DefaultPrinterDriver = trim(left(cBuffer, nPos - 1))
        oPrinter.DefaultPrinterPort = ltrim(substr(cBuffer, nPos+1))
        if right (this.DefaultPrinterPort, 1) = ":"
            this.DefaultPrinterPort = left(this.DefaultPrinterPort, ;
                                    len(this.DefaultPrinterPort)-1)
        endif
    endif
    RETURN oPrinter
    *-- EoP: GetDefaultPrinter


ENDCLASS
*-- EoC: PrinterProperties

CLASS PrnPropsErrors
    this[1] = resource(189,"CRPEWRAP.DLL")
    this[2] = resource(73,"CRPEWRAP.DLL")
    this[3] = resource(190,"CRPEWRAP.DLL")
ENDCLASS


CLASS VdBDevMode
*-----------------------------------------------------------------------
*-- Programmer..: Jim Sare (CIS: 71562,413)
*-- Date........: 08/04/96
*-- Notes.......: Custom version for use with CRWRAPper
*-- Written for.: Visual dBASE v5.5
*-- Rev. History: 08/04/96 - Original
*-- Calls.......: None
* The new VdBDevMode class uses access/assign methods. The methods
* immediately change the associated DEVMODE structure member directly in
* the structure when a new value is passed to the associated method.
* Also, when calling an associated method without a parameter, the
* method will immediately extract the value from the structure member
* and return it.
*
* The Structure method stores/retrieves an entire structure. The other
* methods then act on the individual members in the structure.
* Essentially, send the structure you retrieve from ExtDeviceMode to the
* Structure method. Then manipulate the individual structure members
* with the associated methods. Then retrieve the structure when needed
* with the Structure method.
*
* The methods....
*  dmDeviceName
*  dmSpecVersion
*  dmDriverVersion
*  dmSize
*  dmDriverExtra
* ....essentially do their thing with little if any fanfare.
*
* The method dmFields is PROTECTed. It is used by other methods of the
* class that change the respective field.
*
* The methods....
*  dmOrientation
*  dmPaperSize
*  dmPaperLength
*  dmPaperWidth
*  dmScale
*  dmCopies
*  dmDefaultSource
*  dmPrintQuality
*  dmColor
*  dmDuplex
*  dmYResolution
*  dmTTOption
*
* ....verify that the setting passed is valid for the respective devmode
* field. If so, it stores the value passed in the structure and sets the
* appropriate dmFields bit valid for the respective field. If the
* setting passed is invalid, it does not store the value passed to the
* structure and sets the appropriate dmFields bit invalid for the
* respective field. Therefore, to turn off a specific flag in dmFields,
* send the respective method for the member a value of -99 ie;
* o.DevMode.dmTTOption(-99) will turn off the dmFields flag for
* TTOption. If o.DevMode.dmTTOption() returns -99, then the TTOption
* flag in dmFields is off.
*
* dmDriverData does nothing to validate or massage the driver-specific
* data in any way. It simply stores/retrieves the data on request.
*
* There is no Release method. None is required. All function is derived
* from stock VdB string and bitwise functions.
*
* There is a logout method for debug that when called will display a
* formatted picture of the current structure status in the results pane.
* Passing a true parameter to the logout method will also format and
* display the extended device driver information in hex/ASCII form.
*-----------------------------------------------------------------------
   #define IND_DMSPECVERSION    33
   #define IND_DMDRIVERVERSION  35
   #define IND_DMSIZE           37
   #define IND_DMDRIVEREXTRA    39
   #define IND_DMFIELDS         41
   #define IND_DMORIENTATION    45
   #define IND_DMPAPERSIZE      47
   #define IND_DMPAPERLENGTH    49
   #define IND_DMPAPERWIDTH     51
   #define IND_DMSCALE          53
   #define IND_DMCOPIES         55
   #define IND_DMDEFAULTSOURCE  57
   #define IND_DMPRINTQUALITY   59
   #define IND_DMCOLOR          61
   #define IND_DMDUPLEX         63
   #define IND_DMYRESOLUTION    65
   #define IND_DMTTOPTION       67

   this.pacClassName = "VDBDEVMODE"
   this.paxDEVMODE =   ""

   PROTECT pacClassName, paxDEVMODE, dmFields

   FUNCTION ClassName
      RETURN this.pacClassName

   FUNCTION dmDeviceName(cVal)
      LOCAL cStr
      cStr = cVal
      If (PCount() > 0) .AND. (Len(cVal) => 0)
         If Right(cStr, 1) # Chr(0)
            cStr = cStr + Chr(0)
         EndIf
         this.paxDEVMODE = Stuff(this.paxDEVMODE, 0, Len(cStr), cStr)
      EndIf
      RETURN SubStr(this.paxDEVMODE, 1, At(Chr(0), this.paxDEVMODE) - 1)

   FUNCTION dmSpecVersion(nVal)
      If PCount() > 0
         this.paxDEVMODE = Stuff(this.paxDEVMODE, IND_DMSPECVERSION, 2, STORE16(nVal))
      EndIf
      RETURN MAKEU16(this.paxDEVMODE, IND_DMSPECVERSION)

   FUNCTION dmDriverVersion(nVal)
      If PCount() > 0
         this.paxDEVMODE = Stuff(this.paxDEVMODE, IND_DMDRIVERVERSION, 2, STORE16(nVal))
      EndIf
      RETURN MAKEU16(this.paxDEVMODE, IND_DMDRIVERVERSION)

   FUNCTION dmSize(nVal)
      If (PCount() > 0) .AND. (nVal => 0)
         this.paxDEVMODE = Stuff(this.paxDEVMODE, IND_DMSIZE, 2, STORE16(nVal))
      EndIf
      RETURN MAKEU16(this.paxDEVMODE, IND_DMSIZE)

   FUNCTION dmDriverExtra(nVal)
      If (PCount() > 0) .AND. (nVal => 0)
         this.paxDEVMODE = Stuff(this.paxDEVMODE, IND_DMDRIVEREXTRA, 2, STORE16(nVal))
      EndIf
      RETURN MAKEU16(this.paxDEVMODE, IND_DMDRIVEREXTRA)

   FUNCTION dmFields(nVal)
      If (PCount() > 0) .AND. (nVal => 0)
         this.paxDEVMODE = Stuff(this.paxDEVMODE, IND_DMFIELDS, 4, STORE32(nVal))
      EndIf
      RETURN MAKEU32(this.paxDEVMODE, IND_DMFIELDS)

   FUNCTION dmOrientation(nVal)
      If PCount() > 0
         If (nVal = 1) .OR. (nVal = 2)
            this.paxDEVMODE = Stuff(this.paxDEVMODE, IND_DMORIENTATION, 2, STORE16(nVal))
            this.dmFields(BitOr(this.dmFields(), DM_ORIENTATION))
         ElseIf BitAnd(this.dmFields(), DM_ORIENTATION) > 0
            this.dmFields(BitXOr(this.dmFields(), DM_ORIENTATION))
         EndIf
      EndIf
      RETURN IIf(BitAnd(this.dmFields(), DM_ORIENTATION) > 0,;
                 MAKEU16(this.paxDEVMODE, IND_DMORIENTATION),;
                 -99)

   FUNCTION dmPaperSize(nVal)
      If (PCount() > 0)
         If ((nVal => 1) .AND. (nVal =<41)) .OR. (nVal = 256)
            this.paxDEVMODE = Stuff(this.paxDEVMODE, IND_DMPAPERSIZE, 2, STORE16(nVal))
            this.dmFields(BitOr(this.dmFields(), DM_PAPERSIZE))
         ElseIf BitAnd(this.dmFields(), DM_PAPERSIZE) > 0
            this.dmFields(BitXOr(this.dmFields(), DM_PAPERSIZE))
         EndIf
      EndIf
      RETURN IIf(BitAnd(this.dmFields(), DM_PAPERSIZE) > 0,;
                 MAKEU16(this.paxDEVMODE, IND_DMPAPERSIZE),;
                 -99)

   FUNCTION dmPaperLength(nVal)
      If PCount() > 0
         If nVal => 1
            this.paxDEVMODE = Stuff(this.paxDEVMODE, IND_DMPAPERLENGTH, 2, STORE16(nVal))
            this.dmFields(BitOr(this.dmFields(), DM_PAPERLENGTH))
         ElseIf BitAnd(this.dmFields(), DM_PAPERLENGTH) > 0
            this.dmFields(BitXOr(this.dmFields(), DM_PAPERLENGTH))
         EndIf
      EndIf
      RETURN IIf(BitAnd(this.dmFields(), DM_PAPERLENGTH) > 0,;
                 MAKEU16(this.paxDEVMODE, IND_DMPAPERLENGTH),;
                 -99)

   FUNCTION dmPaperWidth(nVal)
      If PCount() > 0
         If nVal => 1
            this.paxDEVMODE = Stuff(this.paxDEVMODE, IND_DMPAPERWIDTH, 2, STORE16(nVal))
            this.dmFields(BitOr(this.dmFields(), DM_PAPERWIDTH))
         ElseIf BitAnd(this.dmFields(), DM_PAPERWIDTH) > 0
            this.dmFields(BitXOr(this.dmFields(), DM_PAPERWIDTH))
         EndIf
      EndIf
      RETURN IIf(BitAnd(this.dmFields(), DM_PAPERWIDTH) > 0,;
                 MAKEU16(this.paxDEVMODE, IND_DMPAPERWIDTH),;
                 -99)

   FUNCTION dmScale(nVal)
      If PCount() > 0
         If nVal => 1
            this.paxDEVMODE = Stuff(this.paxDEVMODE, IND_DMSCALE, 2, STORE16(nVal))
            this.dmFields(BitOr(this.dmFields(), DM_SCALE))
         ElseIf BitAnd(this.dmFields(), DM_SCALE) > 0
            this.dmFields(BitXOr(this.dmFields(), DM_SCALE))
         EndIf
      EndIf
      RETURN IIf(BitAnd(this.dmFields(), DM_SCALE) > 0,;
                 MAKEU16(this.paxDEVMODE, IND_DMSCALE),;
                 -99)

   FUNCTION dmCopies(nVal)
      If PCount() > 0
         If nVal => 1
            this.paxDEVMODE = Stuff(this.paxDEVMODE, IND_DMCOPIES, 2, STORE16(nVal))
            this.dmFields(BitOr(this.dmFields(), DM_COPIES))
         ElseIf BitAnd(this.dmFields(), DM_COPIES) > 0
            this.dmFields(BitXOr(this.dmFields(), DM_COPIES))
         EndIf
      EndIf
      RETURN IIf(BitAnd(this.dmFields(), DM_COPIES) > 0,;
                 MAKEU16(this.paxDEVMODE, IND_DMCOPIES),;
                 -99)

   FUNCTION dmDefaultSource(nVal)
      If PCount() > 0
         If ((nVal => 1) .AND. (nVal =<14)) .OR. (nVal = 256)
            this.paxDEVMODE = Stuff(this.paxDEVMODE, IND_DMDEFAULTSOURCE, 2, STORE16(nVal))
            this.dmFields(BitOr(this.dmFields(), DM_DEFAULTSOURCE))
         ElseIf BitAnd(this.dmFields(), DM_DEFAULTSOURCE) > 0
            this.dmFields(BitXOr(this.dmFields(), DM_DEFAULTSOURCE))
         EndIf
      EndIf
      RETURN IIf(BitAnd(this.dmFields(), DM_DEFAULTSOURCE) > 0,;
                 MAKEU16(this.paxDEVMODE, IND_DMDEFAULTSOURCE),;
                 -99)

   FUNCTION dmPrintQuality(nVal)
      If PCount() > 0
         If (nVal => -4)
            this.paxDEVMODE = Stuff(this.paxDEVMODE, IND_DMPRINTQUALITY, 2, STORE16(nVal))
            this.dmFields(BitOr(this.dmFields(), DM_PRINTQUALITY))
         ElseIf BitAnd(this.dmFields(), DM_PRINTQUALITY) > 0
            this.dmFields(BitXOr(this.dmFields(), DM_PRINTQUALITY))
         EndIf
      EndIf
      RETURN IIf(BitAnd(this.dmFields(), DM_PRINTQUALITY) > 0,;
                 MAKES16(this.paxDEVMODE, IND_DMPRINTQUALITY),;
                 -99)

   FUNCTION dmColor(nVal)
      If PCount() > 0
         If (nVal = 1) .OR. (nVal = 2)
            this.paxDEVMODE = Stuff(this.paxDEVMODE, IND_DMCOLOR, 2, STORE16(nVal))
            this.dmFields(BitOr(this.dmFields(), DM_COLOR))
         ElseIf BitAnd(this.dmFields(), DM_COLOR) > 0
            this.dmFields(BitXOr(this.dmFields(), DM_COLOR))
         EndIf
      EndIf
      RETURN IIf(BitAnd(this.dmFields(), DM_COLOR) > 0,;
                 MAKEU16(this.paxDEVMODE, IND_DMCOLOR),;
                 -99)

   FUNCTION dmDuplex(nVal)
      If PCount() > 0
         If (nVal => 1) .AND. (nVal =< 3)
            this.paxDEVMODE = Stuff(this.paxDEVMODE, IND_DMDUPLEX, 2, STORE16(nVal))
            this.dmFields(BitOr(this.dmFields(), DM_DUPLEX))
         ElseIf BitAnd(this.dmFields(), DM_DUPLEX) > 0
            this.dmFields(BitXOr(this.dmFields(), DM_DUPLEX))
         EndIf
      EndIf
      RETURN IIf(BitAnd(this.dmFields(), DM_DUPLEX) > 0,;
                 MAKEU16(this.paxDEVMODE, IND_DMDUPLEX),;
                 -99)

   FUNCTION dmYResolution(nVal)
      If PCount() > 0
         If nVal => 1
            this.paxDEVMODE = Stuff(this.paxDEVMODE, IND_DMYRESOLUTION, 2, STORE16(nVal))
            this.dmFields(BitOr(this.dmFields(), DM_YRESOLUTION))
         ElseIf BitAnd(this.dmFields(), DM_YRESOLUTION) > 0
            this.dmFields(BitXOr(this.dmFields(), DM_YRESOLUTION))
         EndIf
      EndIf
      RETURN IIf(BitAnd(this.dmFields(), DM_YRESOLUTION) > 0,;
                 MAKEU16(this.paxDEVMODE, IND_DMYRESOLUTION),;
                 -99)

   FUNCTION dmTTOption(nVal)
      If PCount() > 0
         If (nVal => 1) .AND. (nVal =< 3)
            this.paxDEVMODE = Stuff(this.paxDEVMODE, IND_DMTTOPTION, 2, STORE16(nVal))
            this.dmFields(BitOr(this.dmFields(), DM_TTOPTION))
         ElseIf BitAnd(this.dmFields(), DM_TTOPTION) > 0
            this.dmFields(BitXOr(this.dmFields(), DM_TTOPTION))
         EndIf
      EndIf
      RETURN IIf(BitAnd(this.dmFields(), DM_TTOPTION) > 0,;
                 MAKEU16(this.paxDEVMODE, IND_DMTTOPTION),;
                 -99)

   FUNCTION dmDriverData(cVal)
      If (PCount() > 0) .AND. (Len(cVal) => 0)
         this.paxDEVMODE = Stuff(this.paxDEVMODE, this.dmSize() + 1, Len(cVal), cVal)
         this.dmDriverExtra(Len(cVal))
      EndIf
      RETURN SubStr(this.paxDEVMODE, this.dmSize() + 1, this.dmDriverExtra())

   FUNCTION SizeOf(nVal)
      If (PCount() > 0) .AND. (nVal > 0)
         this.paxDEVMODE = Replicate(Chr(0), nVal)
      EndIf
      RETURN Len(this.paxDEVMODE)

   FUNCTION Structure(xDEVMODE)
      If (PCount() > 0) .AND. (Len(xDEVMODE) => 0)
         this.paxDEVMODE = xDEVMODE
      EndIf
      RETURN this.paxDEVMODE

   PROCEDURE Logout(lDD)

      LOCAL x, xStr, nVal, cAscStr
      ? "DeviceName:    " + this.dmDeviceName()
      ? "SpecVersion:   " + LTrim(Str(this.dmSpecVersion()))
      ? "DriverVersion: " + LTrim(Str(this.dmDriverVersion()))
      ? "Size:          " + LTrim(Str(this.dmSize()))
      ? "DriverExtra:   " + LTrim(Str(this.dmDriverExtra()))
      ? "Fields:        x" + IToH(this.dmFields())
      ? "Orientation:   " + LTrim(Str(this.dmOrientation()))
      ? "PaperSize:     " + LTrim(Str(this.dmPaperSize()))
      ? "PaperLength:   " + LTrim(Str(this.dmPaperLength()))
      ? "PaperWidth:    " + LTrim(Str(this.dmPaperWidth()))
      ? "Scale:         " + LTrim(Str(this.dmScale()))
      ? "Copies:        " + LTrim(Str(this.dmCopies()))
      ? "DefaultSource: " + LTrim(Str(this.dmDefaultSource()))
      ? "PrintQuality:  " + LTrim(Str(this.dmPrintQuality()))
      ? "Color:         " + LTrim(Str(this.dmColor()))
      ? "Duplex:        " + LTrim(Str(this.dmDuplex()))
      ? "YResolution:   " + LTrim(Str(this.dmYResolution()))
      ? "TTOption:      " + LTrim(Str(this.dmTTOption()))
      If lDD
         ? "DriverExtra Data:"
         ?
         x = 1
         y = 1
         xStr = this.dmDriverData()
         cAscStr = " "
         Do
            nVal = MAKE8(xStr, x)
            cAscStr = cAscStr + IIf((nVal < 32) .OR. (nVal > 127), ".", Chr(nVal))
            ?? IToH(nVal, 2) + " "
            If Mod(x, 8) = 0 .OR. x = this.dmDriverExtra()
               ?? cAscStr At 24
               ?
               cAscStr = " "
            EndIf
            x = x + 1
         Until x > this.dmDriverExtra()
      EndIf

ENDCLASS  && VdBDevMode


CLASS PrinterProfiles
*-----------------------------------------------------------------------
*-- Programmer..: Bowen Moursund (CIS: 72662,436)
*-- Date........: 02/28/97
*-- Notes.......: This class will read WIN.INI and parse the [Devices]
*--               section into 2 array objects. The Printers array
*--               is 2 dimensional, with 3 columns as follows:
*--
*--                   Column 1        Column 2       Column 3
*--                 ------------   --------------   -----------
*--                 printer name   printer driver      port
*--
*--               The PrinterList array contains the printers' names
*--               and ports concatenated into a string suitable for
*--               display in a listbox or combobox. The rightmost
*--               characters of each array element are the index
*--               number into the related printer's row in the
*--               Printers array.
*--
*--
*-- Written for.: dBASEWin 5.5
*-- Rev. History: 05/14/95 - Revised to use aGetProfileStrings
*--               03/31/96 - ListString method added, minor revisions
*--               02/28/97 - aGetProfileStrings merged as method
*--                        - Internationalization
*-----------------------------------------------------------------------

  this.Printers = new array(1,3)
  this.PrinterList = new array()
  private aProfileStrings
  ProfileStrings = this.GetProfileStrings("devices")
  class::ParsePrinterProfileStrings(ProfileStrings)


  PROCEDURE ParsePrinterProfileStrings
  parameter a  && ProfileStrings array object reference
  local nDevice, nPort, nEquPos, aComma, nBeg, nEnd, cName, cPort, ;
        cDriver, lFirst, nSubscript
  lFirst = .t.

  for nDevice = 1 to a.size

    nEquPos = at("=",a[nDevice])
    aComma = new array(6)  && limit of 5 ports per device
    aComma[1] = at(",",a[nDevice],1)
    aComma[2] = at(",",a[nDevice],2)
    aComma[3] = at(",",a[nDevice],3)
    aComma[4] = at(",",a[nDevice],4)
    aComma[5] = at(",",a[nDevice],5)
    aComma[6] = 0

    for nPort = 1 to 5
      nBeg = aComma[nPort]+1
      if aComma[nPort+1] >0  && if more than one (or another) port
        nEnd = aComma[nPort+1] - aComma[nPort] - 1
        if .not. lFirst
          agrow(this.Printers,1)
        else
          lFirst = .f.
        endif
        cName = ltrim(trim(substr(a[nDevice], 1, nEquPos-1)))
        cDriver = ltrim(trim(substr(a[nDevice], nEquPos+1, ;
                        (aComma[1]-nEquPos-1) )))
        cPort = ltrim(trim(substr( a[nDevice], nBeg, nEnd )))
        nSubscript = this.PrinterList.Size + 1
        aresize(this.PrinterList, nSubScript)
        this.PrinterList[nSubscript] =  class::ListString(cName, cPort) +;
                                  space(100) + ltrim(str(nSubscript))
        this.Printers[nSubscript,1] = cName
        this.Printers[nSubscript,2] = cDriver
        this.Printers[nSubscript,3] = cPort
      else
        nEnd = 5000
        if .not. lFirst
          agrow(this.Printers,1)
        else
          lFirst = .f.
        endif
        cName = ltrim(trim(substr(a[nDevice], 1, nEquPos-1)))
        cDriver = ltrim(trim(substr(a[nDevice], nEquPos+1, ;
                        (aComma[1]-nEquPos-1) )))
        cPort = ltrim(trim(substr( a[nDevice], nBeg, nEnd )))
        nSubscript = this.PrinterList.Size + 1
        aresize(this.PrinterList, nSubScript)
        this.PrinterList[nSubscript] =  class::ListString(cName, cPort) +;
                                  space(100) + ltrim(str(nSubscript))
        this.Printers[nSubscript,1] = cName
        this.Printers[nSubscript,2] = cDriver
        this.Printers[nSubscript,3] = cPort
        exit
      endif

    endfor

  endfor

  PROCEDURE ListString(cPrinterName, cPrinterPort)
  local cTemp
  cTemp = cPrinterPort
  if right (cTemp, 1) = ":"
      cTemp = substr(cTemp, 1, len(cTemp)-1)
  endif
  RETURN cPrinterName + " on " + cTemp

  PROCEDURE GetProfileStrings(cSection)
  *-- Returns reference to array containing all entries for the
  *-- specified WIN.INI file section
  local cSectionBuffer, cNULL, nBytesUsed, cValueBuffer, nPosition,;
        cValue, cEntry, aArray, cDefault
  cDefault = resource(179,"CRPEWRAP.DLL")
  if type("GetProfileString") <> "FP"
      extern CINT GetProfileString(CSTRING, CSTRING, CSTRING,;
                                   CPTR, CINT) KERNEL
  endif
  cSectionBuffer = space(32000)
  cNULL = 0
  nBytesUsed = GetProfileString(cSection, cNULL, cDefault,;
                                cSectionBuffer, 32000)
  if nBytesUsed = 0
      aArray = new array(1)
      aArray[1] = "[" + cSection + "]" + cDefault
      RETURN aArray
  endif
  cValueBuffer = space(320)
  nPosition = 1
  aArray = new array(0)
  do while nPosition < nBytesUsed
      cEntry = substr(cSectionBuffer, nPosition)
      cEntry = left(cEntry, at(chr(0), cEntry) - 1)
      nPosition = nPosition + len(cEntry) + 1
      GetProfileString(cSection, cEntry, cEntry + " " +;
                       cDefault, cValueBuffer, 320)
      cValue = left(cValueBuffer, at(chr(0), cValueBuffer) - 1)
      aArray.add(cEntry + "=" + cValue)
  enddo
  RETURN aArray
  *-- EoP: GetProfileStrings

ENDCLASS
*-- EoC: PrinterProfiles


FUNCTION GetDefaultPrinter
local oPrinter, cBuffer, nBytesUsed, nPos, cDefault
if type("GetProfileString") <> "FP"
    extern CINT GetProfileString(CSTRING, CSTRING, CSTRING,;
                                 CPTR, CINT) KERNEL
endif
cBuffer = space(100)
*cDefault = resource(179,"CRPEWRAP.DLL")
cDefault = "not found"
nBytesUsed = GetProfileString("windows", "device", cDefault, cBuffer, 254)
cBuffer = ltrim(trim(cBuffer))
if right(cBuffer,1) = chr(0)
    cBuffer = left(cBuffer, len(cBuffer)-1)
endif
oPrinter = new object()
if nBytesUsed = 0
    oPrinter.Name = cDefault
    oPrinter.Driver = cDefault
    oPrinter.Port = cDefault
else
    nPos = at(",", cBuffer)
    oPrinter.Name = trim(left(cBuffer, nPos - 1))
    cBuffer = ltrim(substr(cBuffer, nPos+1))
    nPos = at(",", cBuffer)
    oPrinter.Driver = trim(left(cBuffer, nPos - 1))
    oPrinter.Port = ltrim(substr(cBuffer, nPos+1))
    if right (oPrinter.Port, 1) = ":"
        oPrinter.Port = left(oPrinter.Port, len(oPrinter.Port)-1)
    endif
endif
RETURN oPrinter
*-- EoF: GetDefaultPrinter

*-- EoF: printer.prg
